import logging.config
import logging.handlers

from pathlib import Path
from typing import cast

from hylog import config
from hylog import handlers
from hylog import logger


Config = config.Config()


def get_app_logger(
    app_name: str,
    output_dir: str | Path,
    stdout_level: str | None = None,
) -> logger._AppLogger:
    """Create a logger for the application with the given name and output directory."""
    # TODO: Add support for showing path to module rather than just the name?
    # TODO: Add support to disable file logging
    Config.app.name = app_name
    Config.app.output_dir = Path(output_dir)

    logging.setLoggerClass(logger._AppLogger)

    # Configure all handlers and logging levels into a QueueHandler
    handlers.setup_handlers(
        output_dir=Path(output_dir), name=app_name, stdout_level=stdout_level
    )

    Config.app.initialized = True

    return cast(logger._AppLogger, logging.getLogger(app_name))

def get_logger(
    name: str | None = None,
) -> logger._AppLogger:
    """Create a logger for the application with the given name and output directory."""
    if not Config.app.initialized:
        raise RuntimeError("App logger must be initialized before creating a logger")

    elif name == Config.app.name or name is None:
        return cast(logger._AppLogger ,logging.getLogger(Config.app.name))
    else:
        return cast(logger._AppLogger ,logging.getLogger(Config.app.name).getChild(name))


import logging

from pathlib import Path


class AppConfig:
    name: str | None = None
    output_dir: Path | None = None
    initialized: bool = False

class FileConfig:
    level: int = logging.DEBUG
    max_bytes: int = 3_000_000
    backup_count: int = 3
    mode = "w"
    rotating_suffix = "_rotating.log"
    last_suffix = "_last.log"
    json_suffix = "_json.jsonl"

class StreamConfig:
    level: int = logging.WARNING

class Config:
    file: FileConfig = FileConfig()
    stream: StreamConfig = StreamConfig()
    app: AppConfig = AppConfig()



import time

from pathlib import Path

from hylog import get_app_logger


def test_logger(name: str, output_dir: str | Path) -> None:
    log = get_app_logger(name, output_dir, stdout_level="DEBUG")

    @log.func()
    def test_func_decorator(arg1: str, *, kwarg1: int) -> str:
        return f"arg1: {arg1}, kwarg1: {kwarg1}"

    @log.perf()
    def test_perf_decorator() -> None:
        time.sleep(1)

    def test_logger() -> None:
        log.debug("DEBUG message")
        log.debug("DEBUG message with extra", extra={"extra_key": "extra_value"})
        log.info("INFO message")
        log.warning("WARNING message")
        log.error("ERROR message")
        log.critical("CRITICAL message")

        log.info(Path().cwd())
        log.info(Path().home())

        try:
            1 / 0

        except ZeroDivisionError as e:
            log.exception("Exception occurred", exc_info=e)

    test_func_decorator("test", kwarg1=42)
    test_perf_decorator()

    



if __name__ == "__main__":
    test_logger("hyLog", Path.cwd() / "logs")



import functools
import inspect
import logging
import time

from collections.abc import Callable
from typing import Any


class _AppLogger(logging.Logger):
    """Custom logger class that adds decorator functionality on top of the standard logger."""

    def func(self, level: str | None = None) -> Callable[..., Any]:
        """A decorator that logs the function inputs and outputs at the given level."""
        # Set the level to the log_level if it exists otherwise use the logger.getLevelName
        # to retrieve the corresponding int value for the log level
        _level = getattr(logging, level.upper()) if level else self.level

        def decorator(_func: Callable[..., Any]) -> Any:
            """Print the function signature and return value"""

            @functools.wraps(_func)
            def wrapper_debug(*args: Any, **kwargs: Any) -> Any:
                # Get the function's signature
                func_path = f"{_func.__module__}.{_func.__qualname__}"
                signature = inspect.signature(_func)
                func_line = inspect.getsourcelines(_func)[1]

                # Change the stacklevel to 2 so the log message has the module and lineno of the function call, not the decorator.
                log = functools.partial(self.log, stacklevel = 2)

                log(_level, f"Function: {func_path} (line {func_line})")

                # Map the *args and **kwargs to parameter names
                bound_arguments = signature.bind(*args, **kwargs)
                bound_arguments.apply_defaults()

                log(_level, f"{_func.__name__}() arguments:")

                for name, value in bound_arguments.arguments.items():
                    log(_level, f"{name} = {value}")

                args_repr = [f"{a!r}" for a in args]
                kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
                signature = ", ".join(args_repr + kwargs_repr)
                log(_level, f"Calling: {_func.__name__}({signature})")

                value = _func(*args, **kwargs)
                log(_level, "Returned:")
                log(_level, f"{value!r}")

                return value

            return wrapper_debug

        return decorator

    def perf(self, level: str | None = None) -> Callable[..., Any]:
        """A decorator that logs the time taken by the function to execute at the given level."""
        # Set the level to the log_level if it exists otherwise use the logger.getLevelName
        # to retrieve the corresponding int value for the log level
        _level = getattr(logging, level.upper()) if level else self.level

        def decorator(func: Callable[..., Any]) -> Any:
            """Print the function signature and return value"""

            @functools.wraps(func)
            def wrapper_debug(*args: Any, **kwargs: Any) -> Any:
                start_time = time.perf_counter()
                value = func(*args, **kwargs)
                end_time = time.perf_counter()
                total_time = end_time - start_time

                # Change the stacklevel to 2 so the log message has the module and lineno of the function call, not the decorator.
                log = functools.partial(self.log, stacklevel = 2)
                log(
                    _level, f"{total_time:.4f} seconds for Function {func.__name__}"
                )

                return value

            return wrapper_debug

        return decorator



import atexit
import logging.handlers
import sys

from multiprocessing import Queue
from pathlib import Path

from hylog import config
from hylog import formatters


Config = config.Config()

# File handlers
def _format_output_file_path(*args, **kwargs) -> tuple[Path, str]:
    """Retrieve and validate that the user passed a name and output directory.

    Returns the path to the output file.
    """
    output_dir = kwargs.get("output_dir")
    name = kwargs.get("name")

    if output_dir is None or name is None:
        raise ValueError(
            f"output_dir and name must be provided. Got {output_dir=} and {name=}"
        )

    if not Path(output_dir).is_dir():
        raise FileNotFoundError(f"Output directory {output_dir} does not exist.")

    return Path(output_dir), name

class FileLastRun(logging.FileHandler):
    _formatter: logging.Formatter = formatters.Detailed()
    _mode: str = "w"

    def __init__(self, *args, **kwargs) -> None:
        """
        Logs the last run of the application to a file.

        The file is overwritten each time the application is run.
        """
        output_dir, name = _format_output_file_path(*args, **kwargs)
        file_path = output_dir / (name + Config.file.last_suffix)

        super().__init__(file_path, mode=self._mode)

        self.setLevel(Config.file.level)
        self.setFormatter(self._formatter)


class FileRotating(logging.handlers.RotatingFileHandler):
    _formatter: logging.Formatter = formatters.Detailed()
    # _level: int = logging.DEBUG
    # _suffix = "_rotating.log"

    _max_bytes: int = 3_000_000
    _backup_count: int = 3

    def __init__(self, *args, **kwargs) -> None:
        output_dir, name = _format_output_file_path(*args, **kwargs)
        file_path = output_dir / (name + Config.file.rotating_suffix)

        super().__init__(
            file_path, maxBytes=self._max_bytes, backupCount=self._backup_count
        )

        self.setLevel(Config.file.level)
        self.setFormatter(self._formatter)


class JSONHandler(logging.handlers.RotatingFileHandler):
    _formatter: logging.Formatter = formatters.JSON()

    def __init__(self, *args, **kwargs) -> None:
        output_dir, name = _format_output_file_path(*args, **kwargs)
        file_path = output_dir / (name + Config.file.json_suffix)

        super().__init__(
            file_path, maxBytes=Config.file.max_bytes, backupCount=Config.file.backup_count
        )

        self.setLevel(Config.file.level)
        self.setFormatter(self._formatter)


# Stream handlers
class StandardOutput(logging.StreamHandler):
    _formatter: logging.Formatter = formatters.Simple()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(stream=sys.stdout)
        user_level = kwargs.get("stdout_level")
        if user_level is not None:
            _stdout_level = getattr(logging, user_level.upper())
        else:
            _stdout_level = Config.stream.level

        self.setLevel(_stdout_level)
        self.setFormatter(self._formatter)


# Queue handlers
class QueueHandler(logging.handlers.QueueHandler):
    def __init__(self, queue: Queue) -> None:
        super().__init__(queue)


class QueueListener(logging.handlers.QueueListener):
    def __init__(self, queue, *handlers: logging.Handler) -> None:
        super().__init__(queue, *handlers, respect_handler_level=True)


def setup_handlers(*args, **kwargs) -> None:
    """Instantiae and add all handlers to the QueueListener/Handler and configure logger."""
    name = kwargs.get("name", "default_logger")
    if name is None:
        raise ValueError("Logger name must be provided")

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    log_queue = Queue(-1)
    queue_handler = QueueHandler(log_queue)

    logger.addHandler(queue_handler)

    handler_classes = [
        StandardOutput,
        FileLastRun,
        FileRotating,
        JSONHandler,
    ]

    queue_listener = QueueListener(
        log_queue,
        *[handler(*args, **kwargs) for handler in handler_classes],
    )

    queue_listener.start()
    atexit.register(queue_listener.stop)



import datetime as dt
import json
import logging

from typing import ClassVar
from typing import override


LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


class Simple(logging.Formatter):
    _fmt: str = "%(asctime)s - [%(levelname)-8s] %(module)s:%(lineno)03d | %(message)s"
    _datefmt: str = "T%H:%M:%S"

    def __init__(self) -> None:
        super().__init__(fmt=self._fmt, datefmt=self._datefmt)


class Detailed(logging.Formatter):
    _fmt: str = "%(asctime)s [%(levelname)-8s]  %(pathname)s:%(lineno)03d %(module)s %(funcName)s | %(message)s"
    _datefmt: str = "%Y-%m-%dT%H:%M:%S"

    def __init__(self) -> None:
        super().__init__(fmt=self._fmt, datefmt=self._datefmt)


class JSON(logging.Formatter):
    _fmt_keys: ClassVar[dict[str, str]] = {
        "level": "levelname",
        "message": "message",
        "timestamp": "timestamp",
        "logger": "name",
        "module": "module",
        "function": "funcName",
        "line": "lineno",
        "thread_name": "threadName",
    }

    def __init__(
        self,
        *,
        fmt_keys: dict[str, str] | None = None,
    ) -> None:
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else self._fmt_keys

    @override
    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord) -> dict[str, str]:
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created, tz=dt.UTC
            ).isoformat(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: msg_val
            if (msg_val := always_fields.pop(val, None)) is not None
            else getattr(record, val)
            for key, val in self.fmt_keys.items()
        }

        message.update(always_fields)

        for key, val in record.__dict__.items():
            if key not in LOG_RECORD_BUILTIN_ATTRS:
                message[key] = val

        return message


class NonErrorFilter(logging.Filter):
    @override
    def filter(self, record: logging.LogRecord) -> bool | logging.LogRecord:
        return record.levelno <= logging.ERROR
